from django.dispatch import receiver
from django.test.signals import setting_changed
from wagtail.permission_policies.collections import CollectionOwnershipPermissionPolicy

from wideo import get_video_model
from wideo.models import Video

permission_policy = None


class VideosPermissionPolicyGetter:
    def __get__(self, obj, objtype=None):
        return permission_policy


def set_permission_policy():
    global permission_policy
    permission_policy = CollectionOwnershipPermissionPolicy(
        get_video_model(), auth_model=Video, owner_field_name="uploaded_by_user"
    )


@receiver(setting_changed)
def update_permission_policy(signal, sender, setting, **kwargs):
    if setting == "WIDEO_VIDEO_MODEL":
        set_permission_policy()


set_permission_policy()



from celery import shared_task
from django.conf import settings
from django.db.transaction import atomic


def is_celery_present() -> bool:
    """
    Checks is Celery is supposed to be used. Reasonably, checking `CELERY_BROKER_URL`
    should be a reliable indicator of whether the current project uses Celery or not.
    """
    return bool(getattr(settings, "CELERY_BROKER_URL", False))


@shared_task
def delete_orphan_uploaded_videos():
    from .models import delete_orphan_uploaded_videos

    delete_orphan_uploaded_videos()


@shared_task
@atomic
def encode_video(video_id: int):
    from .ffmpeg import encode_video

    encode_video(video_id)



from typing import Any

from django.template.defaulttags import register


@register.simple_tag
def settings(name: str, default: Any = None) -> Any:
    from django.conf import settings

    return getattr(settings, name, default)



from wagtail.blocks import BooleanBlock, StructBlock
from wagtail.snippets.blocks import SnippetChooserBlock

from wideo import get_video_model_string


class VideoBlock(StructBlock):
    video = SnippetChooserBlock(target_model=get_video_model_string())
    autoplay = BooleanBlock(required=False)
    controls = BooleanBlock(required=False)
    muted = BooleanBlock(required=False)

    class Meta:
        label = "Video"
        template = "wideo/blocks/video.html"



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured


def get_video_model_string():
    """
    Get the dotted ``app.Model`` name for the video model as a string, just like
    ``get_image_model_string()`` from wagtailimages.
    """
    return getattr(settings, "WIDEO_VIDEO_MODEL", "wideo.Video")


def get_video_model():
    """
    Get the image model from the ``WAGTAILIMAGES_IMAGE_MODEL`` setting, just like
    ``get_image_model()`` from wagtailimages.
    """
    from django.apps import apps

    model_string = get_video_model_string()

    try:
        return apps.get_model(model_string, require_ready=False)
    except ValueError:
        raise ImproperlyConfigured(
            "WIDEO_VIDEO_MODEL must be of the form 'app_label.model_name'"
        )
    except LookupError:
        raise ImproperlyConfigured(
            "WIDEO_VIDEO_MODEL refers to model '%s' that has not been installed"
            % model_string
        )


def get_render_model_string():
    return getattr(settings, "WIDEO_RENDER_MODEL", "wideo.Render")


def get_render_model():
    from django.apps import apps

    model_string = get_render_model_string()

    try:
        return apps.get_model(model_string, require_ready=False)
    except ValueError:
        raise ImproperlyConfigured(
            "WIDEO_RENDER_MODEL must be of the form 'app_label.model_name'"
        )
    except LookupError:
        raise ImproperlyConfigured(
            "WIDEO_RENDER_MODEL refers to model '%s' that has not been installed"
            % model_string
        )



import os
from uuid import uuid4

from django.db.models import Model
from django.utils.text import slugify


def upload_to(instance: Model, filename: str) -> str:
    """
    We want to make sure to upload the file to a place that isn't going to be
    problematic.

    - In order to limit the number of files in a single directory, we use a
      random UUID in order to create a structure of directories (because most
      storage solutions have limits on the amount of files per folder).
    - The file name should be slugified (using Django's function) to make sure
      there is no problem with special characters.
    - The extension is also going to be normalized and common synonyms are
      going to be replaced by the most common one (like m4v -> mp4).

    Parameters
    ----------
    instance
        An instance
    filename
        Name of the file we're uploading
    """

    file_root, file_ext = os.path.splitext(filename)
    file_ext = file_ext.lower()

    ext_map = {".m4v": ".mp4"}

    file_ext = ext_map.get(file_ext, file_ext)
    slug = slugify(file_root)

    uuid_str = str(uuid4())
    dir_structure = os.path.join(uuid_str[:2], uuid_str[2:4], uuid_str[4:6])

    new_filename = f"{slug}{file_ext}"
    return os.path.join(__package__, dir_structure, new_filename)



from enum import Enum

from django.conf import settings


class Codec(str, Enum):
    """Available code types."""

    H264 = "H264"
    AV1 = "AV1"


# FFMPEG arguments to be used for each specific codec available
CODEC_ARGS = {
    Codec.H264: ["-c:v", "libx264", "-movflags", "faststart"],
    Codec.AV1: ["-c:v", "libaom-av1", "-b:v", "0"],
}

# Video file extension for each codec type
CODEC_EXTENSION = {Codec.H264: "mp4", Codec.AV1: "webm"}

# Mapping between the quality of a video and the number of pixels to encode
VIDEO_QUALITY_TO_PIXELS = {"4K": "3840", "1080P": "1920", "720P": "1280"}


def build_ffmpeg_codec_args(codec: Codec, quality: str) -> list[str]:
    """
    Given a specific codec and the quality of a video, build the arguments to
    be used in a ffmpeg command to encode the video.
    """

    # Default to 1080P if the specified quality is not correct
    if quality not in VIDEO_QUALITY_TO_PIXELS:
        quality = "1080P"

    pixels = VIDEO_QUALITY_TO_PIXELS[quality]
    extra_args = CODEC_ARGS[codec]

    return [
        "-vf",
        f"\"scale='min({pixels},iw)':min'({pixels},ih)':force_original_aspect_ratio=decrease\"",
        "-preset",
        "veryslow",
        *extra_args,
    ]


def build_preset(codec: Codec, quality: str) -> dict:
    """
    Build a preset object for the given codec and video quality.
    """
    return {
        "ffmpeg_flags": build_ffmpeg_codec_args(codec, quality),
        "codec": f'video/{CODEC_EXTENSION[codec]}; codecs="avc1.42E01E, mp4a.40.2"',
        "extension": CODEC_EXTENSION[codec],
    }


def get_presets() -> dict:
    """
    Retrieve the map of available presets from the Django settings if specified.
    Fallback to the default list of presets.
    """
    presets: dict[str, str | dict] = getattr(settings, "WIDEO_PRESETS", DEFAULT_PRESETS)

    return {
        # If a preset is the name of a default preset, get the preset dict from there
        name: (KNOWN_PRESETS.get(preset) if isinstance(preset, str) else preset)
        for name, preset in presets.items()
    }


KNOWN_PRESETS = {
    "4K_H264": build_preset(Codec.H264, "4K"),
    "1080_H264": build_preset(Codec.H264, "1080P"),
    "1080P_H264": build_preset(Codec.H264, "1080P"),
    "720_H264": build_preset(Codec.H264, "720P"),
    "720P_H264": build_preset(Codec.H264, "720P"),
    "4K_AV1": build_preset(Codec.AV1, "4K"),
    "1080_AV1": build_preset(Codec.AV1, "1080P"),
    "1080P_AV1": build_preset(Codec.AV1, "1080P"),
    "720_AV1": build_preset(Codec.AV1, "720P"),
    "720P_AV1": build_preset(Codec.AV1, "720P"),
}

DEFAULT_PRESETS = {
    name: preset for name, preset in KNOWN_PRESETS.items() if name.endswith("H264")
}



# Generated by Django 4.1.9 on 2023-07-06 16:05
import django.db.models.deletion
from django.apps.registry import Apps
from django.db import migrations, models

import wideo.storage


def up(apps: Apps, *args, **kwargs):
    pass


def down(apps: Apps, *args, **kwargs):
    UploadedVideo = apps.get_model("wideo", "UploadedVideo")

    if UploadedVideo.objects.count() > 0:
        raise RuntimeError(
            "UploadedVideo instances must be purged prior to downgrading, as fields without default values must be inserted back"
        )


class Migration(migrations.Migration):
    dependencies = [
        ("wideo", "0002_lock"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="uploadedvideo",
            name="duration",
        ),
        migrations.RemoveField(
            model_name="uploadedvideo",
            name="file",
        ),
        migrations.RemoveField(
            model_name="uploadedvideo",
            name="frame_count",
        ),
        migrations.RemoveField(
            model_name="uploadedvideo",
            name="frames_per_second",
        ),
        migrations.RemoveField(
            model_name="uploadedvideo",
            name="height",
        ),
        migrations.RemoveField(
            model_name="uploadedvideo",
            name="mime",
        ),
        migrations.RemoveField(
            model_name="uploadedvideo",
            name="width",
        ),
        migrations.CreateModel(
            name="UploadedVideoChunk",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "index",
                    models.IntegerField(
                        help_text="The position of the chunk in the uploaded video",
                        verbose_name="index",
                    ),
                ),
                (
                    "file",
                    models.FileField(
                        help_text="The uploaded chunk",
                        upload_to=wideo.storage.upload_to,
                        verbose_name="file",
                    ),
                ),
                (
                    "video",
                    models.ForeignKey(
                        help_text="The video the chunk is part of",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="chunks",
                        to="wideo.uploadedvideo",
                        verbose_name="video",
                    ),
                ),
            ],
            options={
                "unique_together": {("video", "index")},
            },
        ),
        migrations.RunPython(up, down),
    ]



from django.core.management import BaseCommand
from django.utils.translation import gettext_lazy as _

from ...models import delete_orphan_uploaded_videos


class Command(BaseCommand):
    help = _("Deletes any unused user uploaded video")

    def handle(self, *args, **options):
        delete_orphan_uploaded_videos()



from django.db import models

from wideo import get_video_model_string


class VideoField(models.ForeignKey):
    def __init__(self, *args, **kwargs):
        kwargs["to"] = get_video_model_string()
        kwargs["on_delete"] = models.SET_NULL if kwargs.get("null") else models.PROTECT
        super().__init__(*args, **kwargs)



import functools
from datetime import timedelta
from typing import Any, Callable

from django.conf import settings
from django.db import models
from django.db.transaction import atomic
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from taggit.managers import TaggableManager
from wagtail.models import CollectionMember
from wagtail.search import index
from wagtail.snippets.models import register_snippet

from . import get_render_model, get_video_model
from .storage import upload_to


def delete_orphan_uploaded_videos():
    """
    Delete all the instances of UploadedVideo that are not related to any Video
    anymore (result of changing the video file while editing a Video).
    """
    used_ids = get_video_model().objects.values("upload_id")
    UploadedVideo.objects.exclude(id__in=used_ids).filter(
        created_at__lt=now() - timedelta(days=1)
    ).delete()


def lock(name: str) -> Callable:
    def decorator(func: Callable) -> Callable:
        @functools.wraps(func)
        def wrapper(*args, **kwargs) -> Any:
            with atomic():
                Lock.objects.select_for_update().get_or_create(name=name)
                return func(*args, **kwargs)

        return wrapper

    return decorator


class TimestampedModel(models.Model):
    class Meta:
        abstract = True

    created_at = models.DateTimeField(
        verbose_name=_("created at"), auto_now_add=True, db_index=True
    )


class UserUpload(models.Model):
    class Meta:
        abstract = True

    uploaded_by_user = models.ForeignKey(
        to=settings.AUTH_USER_MODEL,
        verbose_name=_("uploaded by user"),
        null=True,
        blank=True,
        editable=False,
        on_delete=models.SET_NULL,
    )
    uploaded_by_user.wagtail_reference_index_ignore = True


class RemoteVideoFile(models.Model):
    class Meta:
        abstract = True

    INFORMATION_FIELDS = (
        "mime",
        "duration",
        "width",
        "height",
        "frames_per_second",
        "frame_count",
    )

    file = models.FileField(
        upload_to=upload_to,
        verbose_name=_("file"),
        help_text=_("The uploaded video file"),
    )
    mime = models.CharField(
        max_length=100,
        verbose_name=_("mime type"),
        help_text=_("MIME type of the video"),
    )
    duration = models.FloatField(
        verbose_name=_("duration in seconds"),
        help_text=_("Duration of the video in seconds"),
    )
    width = models.IntegerField(
        verbose_name=_("original width"),
        help_text=_("Width of the video in pixels"),
    )
    height = models.IntegerField(
        verbose_name=_("original height"),
        help_text=_("Height of the video in pixels"),
    )
    frames_per_second = models.DecimalField(
        verbose_name=_("original FPS"),
        help_text=_("FPS of the video"),
        max_digits=5,
        decimal_places=2,
    )
    frame_count = models.IntegerField(
        verbose_name=_("original frame count"),
        help_text=_("Number of frames in the video"),
    )


class UploadedVideo(TimestampedModel, UserUpload):
    pass


class UploadedVideoChunk(models.Model):
    class Meta:
        unique_together = ("video", "index")

    video = models.ForeignKey(
        to=UploadedVideo,
        on_delete=models.CASCADE,
        related_name="chunks",
        verbose_name=_("video"),
        help_text=_("The video the chunk is part of"),
    )
    index = models.IntegerField(
        verbose_name=_("index"),
        help_text=_("The position of the chunk in the uploaded video"),
    )
    file = models.FileField(
        upload_to=upload_to,
        verbose_name=_("file"),
        help_text=_("The uploaded chunk"),
    )


class AbstractVideo(index.Indexed, CollectionMember, TimestampedModel, UserUpload):
    class Meta:
        abstract = True

    class ProcessStatus(models.TextChoices):
        pending = "pending"
        processing = "processing"
        success = "success"
        failed = "failed"

    title = models.CharField(max_length=255, verbose_name=_("title"))
    upload = models.ForeignKey(
        to=UploadedVideo,
        on_delete=models.PROTECT,
        verbose_name=_("file"),
        help_text=_(
            "The uploaded video file; its resolution should be a standard one to avoid issues (e.g. 1280x720, 1920x1080...)"
        ),
    )
    status = models.CharField(
        max_length=max(len(x) for x, _ in ProcessStatus.choices),
        choices=ProcessStatus.choices,
        default=ProcessStatus.pending,
        verbose_name=_("status"),
        help_text=_(
            "The status of the video processing (pending means nothing "
            "happens on this video yet, processing means the task is running, "
            "success means a render is available and failed means the task "
            "failed)"
        ),
    )
    tags = TaggableManager(help_text=None, blank=True, verbose_name=_("tags"))
    search_fields = CollectionMember.search_fields + [
        index.SearchField("title", boost=10),
        index.AutocompleteField("title"),
        index.FilterField("title"),
        index.RelatedFields(
            "tags",
            [
                index.SearchField("name", boost=10),
                index.AutocompleteField("name"),
            ],
        ),
        index.FilterField("uploaded_by_user"),
    ]
    admin_form_fields = (
        "title",
        "upload",
        "tags",
    )

    @property
    def rendered(self) -> list:
        return [
            {
                "url": render.file.url,
                **{
                    field: getattr(render, field)
                    for field in RemoteVideoFile.INFORMATION_FIELDS
                },
            }
            for render in get_render_model().objects.filter(video=self)
        ]

    def __str__(self):
        return self.title


@register_snippet
class Video(AbstractVideo):
    pass


class AbstractRender(TimestampedModel, RemoteVideoFile):
    """
    If a video as processed correctly, a render is created with pointers to
    all the outputs of the rendering process.
    """

    class Meta:
        abstract = True


class Render(AbstractRender):
    video = models.ForeignKey(
        to=Video,
        on_delete=models.CASCADE,
        verbose_name=_("video"),
        help_text=_("The video that was rendered"),
    )


class Lock(models.Model):
    name = models.TextField(unique=True)



from django import forms

from .models import Video
from .widgets import VideoUploadWidget


class BaseVideoForm(forms.ModelForm):
    class Meta:
        model = Video
        fields = ["title", "upload", "tags"]


class VideoForm(BaseVideoForm):
    class Meta(BaseVideoForm.Meta):
        widgets = {"upload": VideoUploadWidget()}



# Generated by Django 4.1.9 on 2024-01-23 12:18

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wideo", "0003_uploadedvideochunk"),
    ]

    operations = [
        migrations.AlterField(
            model_name="video",
            name="upload",
            field=models.ForeignKey(
                help_text="The uploaded video file; its resolution should be a standard one to avoid issues",
                on_delete=django.db.models.deletion.PROTECT,
                to="wideo.uploadedvideo",
                verbose_name="file",
            ),
        ),
    ]



import subprocess
import uuid
from os import makedirs
from os.path import join
from shutil import rmtree
from subprocess import run
from typing import BinaryIO, Optional

import magic
from django.conf import settings
from django.core.files.base import File
from django.core.files.uploadedfile import InMemoryUploadedFile, TemporaryUploadedFile
from django.db.transaction import atomic

from . import get_render_model, get_video_model
from .codecs import get_presets
from .exceptions import ConfigurationError, InvalidVideoFile
from .models import AbstractRender, AbstractVideo, RemoteVideoFile, lock


def compute_division(division: str) -> Optional[float]:
    if division == "N/A":
        return None

    a, b = division.split("/")
    return float(a) / float(b)


def try_round(n: Optional[float]) -> Optional[float]:
    return round(n, 2) if n is not None else n


def get_video_info(file: BinaryIO) -> dict:
    """
    Uses ffprobe to retrieve some basic information about a video file (size,
    duration...)
    """
    data = None

    if isinstance(file, InMemoryUploadedFile):
        filename = "-"
        position = file.tell()
        file.seek(0)
        data = file.read()
        file.seek(position)
        mime = magic.from_buffer(data, mime=True)
    else:
        filename = (
            file.temporary_file_path()
            if isinstance(file, TemporaryUploadedFile)
            else file.name
        )
        mime = magic.from_file(filename, mime=True)

    ffprobe = run(
        [
            "ffprobe",
            "-select_streams",
            "v:0",
            "-show_entries",
            "stream=width,height,nb_frames,avg_frame_rate:format=duration",
            "-of",
            "default=noprint_wrappers=1",
            filename,
        ],
        input=data,
        capture_output=True,
    )

    info = {
        key: try_round(compute_division(value) if "/" in value else float(value))
        for key, value in (line.split("=") for line in ffprobe.stdout.decode().split())
    }

    if ffprobe.returncode:
        raise InvalidVideoFile

    if not info["nb_frames"]:
        info["nb_frames"] = round(info["avg_frame_rate"] * info["duration"])

    info["mime"] = mime
    # Rename some keys to fit the models used in wideo
    info["frame_count"] = info["nb_frames"]
    info["frames_per_second"] = info["avg_frame_rate"]
    del info["nb_frames"]
    del info["avg_frame_rate"]
    return info


def encode_video(video_id: int):
    """
    Encodes a specific video with whichever presets have been specified in the settings.
    """

    wideo_working_dir = getattr(settings, "WIDEO_WORKING_DIR")
    task_working_dir = join(wideo_working_dir, str(uuid.uuid4()))
    makedirs(task_working_dir)

    if not wideo_working_dir:
        raise ConfigurationError

    with atomic():
        # Before the actual encoding part, flag the video as being processed
        if (
            video := get_video_model()
            .objects.select_for_update()
            .filter(id=video_id)
            .first()
        ):
            video.status = AbstractVideo.ProcessStatus.processing
            video.save()
            # Since we are going to create new renders, delete any previous ones
            get_render_model().objects.filter(video=video).delete()
        else:
            return

    # Encapsulating everything in a separate function makes it easier to try/except, and
    # thus to make sure the status of the video is always changed at the end.
    # noinspection PyBroadException
    try:
        get_video_model().objects.filter(id=video_id).update(
            status=AbstractVideo.ProcessStatus.success
            if encode_video_impl(video, task_working_dir)
            else AbstractVideo.ProcessStatus.failed
        )
    except Exception:
        get_video_model().objects.filter(id=video_id).update(
            status=AbstractVideo.ProcessStatus.failed
        )
        raise
    finally:
        rmtree(task_working_dir)


@lock("encode_video")
def encode_video_impl(video: AbstractVideo, working_dir: str) -> bool:
    """
    The actual implementation of the video encoding using ffmpeg. Return `True` if the
    encoding went down as intended, or `False` if something went wrong.
    """

    presets_map = get_presets()
    renders = {}
    ffmpegs = []

    def abort_everything():
        """
        If something goes wrong during processing (invalid file...), stop everything and
        ensure we don't keep any unused Render objects.
        """
        for f in ffmpegs:
            f.kill()

        for r in renders.values():
            r.delete()

    def get_render_temp_file(r: AbstractRender) -> str:
        """
        Creates a simple unique path to a temporary file, used for storing outputs of
        ffmpeg.
        """
        return join(working_dir, str(r.id))

    # The input file must be written to the disk so that ffmpeg can access it randomly.
    # Streaming the file directly to ffmpeg's stdin would be simpler, but ffmpeg would
    # then not always be able to determine what type of file is its input.

    input_file_path = join(working_dir, "input")

    with open(input_file_path, "wb") as input_file:
        for chunk in video.upload.chunks.order_by("index"):
            with chunk.file.open("rb") as chunk_file:
                input_file.write(chunk_file.read())

    for preset_label, preset in presets_map.items():
        # Fill the render with bogus data until we can get info from the encoded videos
        render = get_render_model().objects.create(
            video=video,
            mime="",
            duration=0,
            width=0,
            height=0,
            frames_per_second=0,
            frame_count=0,
        )
        renders[preset_label] = render
        flags = preset["ffmpeg_flags"]
        command = f"ffmpeg -y -i {input_file_path} -f {preset['extension']} {' '.join(flags)} {get_render_temp_file(render)}"
        ffmpeg = subprocess.Popen(command, shell=True)

        # If parallel work is disabled, wait for each ffmpeg process synchronously to
        # avoid having more than one running at once, and panic if any of them fails.
        # If parallel work is enabled, store the processes to check for their result
        # later on.
        if getattr(settings, "WIDEO_PARALLEL_WORK", False):
            ffmpegs.append(ffmpeg)
        elif ffmpeg.wait():
            abort_everything()
            return False

    # When parallel work is enabled, `ffmpegs` will contain a list with all ffmpeg
    # process that we should wait for. As for when parallel work is disabled, if any of
    # them has failed, panic and abort everything.
    for ffmpeg in ffmpegs:
        if ffmpeg.wait():
            abort_everything()
            return False

    # After all videos have been correctly generated, we just need to write them into
    # their respective Render objects
    for preset_label, preset in presets_map.items():
        render = renders[preset_label]

        with open(get_render_temp_file(render), "rb") as file:
            name = f"{render.video.title}_{preset_label}.{preset['extension']}"
            uploaded_file = File(file)
            # Saving the file this way is necessary; `render.file = File(file, name)`
            # does not work
            render.file.save(name, uploaded_file, save=True)

            # Don't forget to get info from the generated videos, and to store it in the
            # previously created Render objects
            video_info = get_video_info(file)

        for field in RemoteVideoFile.INFORMATION_FIELDS:
            setattr(render, field, video_info[field])

        render.save()

    return True



from django.db.models.signals import post_delete, post_save, pre_delete, pre_save
from django.dispatch import receiver

from . import get_render_model, get_video_model
from .models import AbstractRender, AbstractVideo, UploadedVideoChunk
from .tasks import encode_video, is_celery_present


@receiver(pre_delete, sender=UploadedVideoChunk)
def on_uploaded_video_chunk_pre_delete(instance: UploadedVideoChunk, *args, **kwargs):
    instance.file.delete(save=False)


@receiver(post_delete, sender=get_video_model())
def on_video_post_delete(instance: AbstractVideo, *args, **kwargs):
    instance.upload.delete()


@receiver(pre_save, sender=get_video_model())
def on_video_pre_save(instance: AbstractVideo, *args, **kwargs):
    """
    Whenever a new uploaded video is assigned to a video, encode that video file
    and generate the new renders.
    """
    if old_instance := get_video_model().objects.filter(id=instance.id).first():
        if old_instance.upload_id != instance.upload_id and is_celery_present():
            encode_video.delay(video_id=instance.id)


@receiver(post_save, sender=get_video_model())
def on_video_post_save(instance: AbstractVideo, created: bool, *args, **kwargs):
    """
    Whenever a video is created, encode the video file and generate the renders.
    """
    if created and is_celery_present():
        encode_video.delay(video_id=instance.id)


@receiver(pre_delete, sender=get_render_model())
def on_render_pre_delete(instance: AbstractRender, *args, **kwargs):
    instance.file.delete(save=False)



from django.urls import path

import wideo.views

app_name = "wideo"
urlpatterns = [
    path("", wideo.views.IndexView.as_view(), name="index"),
    path(
        "results/",
        wideo.views.ListingResultsView.as_view(),
        name="listing_results",
    ),
    path("add/", wideo.views.add, name="add"),
    path("<int:video_id>/", wideo.views.edit, name="edit"),
    path("<int:video_id>/delete/", wideo.views.delete, name="delete"),
    path("upload/prepare/", wideo.views.upload_prepare, name="upload-prepare"),
    path("upload/chunk/", wideo.views.upload_chunk, name="upload-chunk"),
    path(
        "get_uploaded_video_render/<int:uploaded_video_id>/",
        wideo.views.get_uploaded_video_render,
        name="get-uploaded-video-render",
    ),
]



import json

from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.core.paginator import Paginator
from django.db.transaction import atomic
from django.http import HttpRequest, HttpResponse
from django.http.multipartparser import MultiPartParser
from django.shortcuts import get_object_or_404, redirect
from django.template.response import TemplateResponse
from django.urls import NoReverseMatch, reverse
from django.utils.decorators import method_decorator
from django.utils.http import urlencode
from django.utils.translation import gettext as _
from django.utils.translation import gettext_lazy, ngettext
from django.views.generic import TemplateView
from rest_framework import status
from wagtail.admin import messages
from wagtail.admin.auth import PermissionPolicyChecker
from wagtail.admin.forms.search import SearchForm
from wagtail.admin.models import popular_tags_for_model
from wagtail.admin.utils import get_valid_next_url_from_request
from wagtail.admin.views import generic
from wagtail.models import Collection
from wagtail.search.backends import get_search_backend

from . import get_render_model, get_video_model
from .forms import BaseVideoForm, VideoForm
from .models import UploadedVideo, UploadedVideoChunk
from .permissions import permission_policy

permission_checker = PermissionPolicyChecker(permission_policy)

INDEX_PAGE_SIZE = getattr(settings, "WIDEO_INDEX_PAGE_SIZE", 30)
USAGE_PAGE_SIZE = getattr(settings, "WIDEO_USAGE_PAGE_SIZE", 20)


class BaseListingView(TemplateView):
    ENTRIES_PER_PAGE_CHOICES = sorted({10, 30, 60, 100, 250, INDEX_PAGE_SIZE})
    ORDERING_OPTIONS = {
        "-created_at": _("Newest"),
        "created_at": _("Oldest"),
        "title": _("Title: (A -> Z)"),
        "-title": _("Title: (Z -> A)"),
        "file_size": _("File size: (low to high)"),
        "-file_size": _("File size: (high to low)"),
    }
    default_ordering = "-created_at"

    @method_decorator(permission_checker.require_any("add", "change", "delete"))
    def get(self, request):
        return super().get(request)

    def get_num_entries_per_page(self):
        entries_per_page = self.request.GET.get("entries_per_page", INDEX_PAGE_SIZE)
        try:
            entries_per_page = int(entries_per_page)
        except ValueError:
            entries_per_page = INDEX_PAGE_SIZE
        if entries_per_page not in self.ENTRIES_PER_PAGE_CHOICES:
            entries_per_page = INDEX_PAGE_SIZE

        return entries_per_page

    def get_valid_orderings(self):
        return self.ORDERING_OPTIONS

    def get_ordering(self):
        # TODO: remove this method when this view will be based on the
        # generic model index view from wagtail.admin.views.generic.models.IndexView
        ordering = self.request.GET.get("ordering")
        if ordering is None or ordering not in self.get_valid_orderings():
            ordering = self.default_ordering
        return ordering

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        # Get ordering
        ordering = self.get_ordering()

        # Get videos (filtered by user permission and ordered by `ordering`)
        videos = (
            permission_policy.instances_user_has_any_permission_for(
                self.request.user, ["change", "delete"]
            )
            .order_by(ordering)
            .select_related("collection")
            # .prefetch_renditions("max-165x165")
        )

        # Filter by collection
        self.current_collection = None
        collection_id = self.request.GET.get("collection_id")
        if collection_id:
            try:
                self.current_collection = Collection.objects.get(id=collection_id)
                videos = videos.filter(collection=self.current_collection)
            except (ValueError, Collection.DoesNotExist):
                pass

        # Search
        query_string = None
        if "q" in self.request.GET:
            self.form = SearchForm(self.request.GET, placeholder=_("Search videos"))
            if self.form.is_valid():
                query_string = self.form.cleaned_data["q"]
                if query_string:
                    search_backend = get_search_backend()
                    videos = search_backend.autocomplete(query_string, videos)
        else:
            self.form = SearchForm(placeholder=_("Search videos"))

        # Filter by tag
        self.current_tag = self.request.GET.get("tag")
        if self.current_tag:
            try:
                videos = videos.filter(tags__name=self.current_tag)
            except AttributeError:
                self.current_tag = None

        entries_per_page = self.get_num_entries_per_page()
        paginator = Paginator(videos, per_page=entries_per_page)
        videos = paginator.get_page(self.request.GET.get("p"))

        next_url = reverse("wideo:index")
        request_query_string = self.request.META.get("QUERY_STRING")
        if request_query_string:
            next_url += "?" + request_query_string

        context.update(
            {
                "videos": videos,
                "query_string": query_string,
                "is_searching": bool(query_string),
                "next": next_url,
                "entries_per_page": entries_per_page,
                "ENTRIES_PER_PAGE_CHOICES": self.ENTRIES_PER_PAGE_CHOICES,
                "current_ordering": ordering,
                "ORDERING_OPTIONS": self.ORDERING_OPTIONS,
            }
        )

        return context


class IndexView(BaseListingView):
    template_name = "wideo/videos/index.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        collections = permission_policy.collections_user_has_any_permission_for(
            self.request.user, ["add", "change"]
        )
        if len(collections) < 2:
            collections = None

        Video = get_video_model()

        context.update(
            {
                "search_form": self.form,
                "popular_tags": popular_tags_for_model(Video),
                "current_tag": self.current_tag,
                "collections": collections,
                "current_collection": self.current_collection,
                "user_can_add": permission_policy.user_has_permission(
                    self.request.user, "add"
                ),
                "app_label": Video._meta.app_label,
                "model_name": Video._meta.model_name,
            }
        )
        return context


class ListingResultsView(BaseListingView):
    template_name = "wideo/videos/results.html"


class DeleteView(generic.DeleteView):
    model = get_video_model()
    pk_url_kwarg = "video_id"
    permission_policy = permission_policy
    permission_required = "delete"
    header_icon = "video"
    template_name = "wideo/videos/confirm_delete.html"
    usage_url_name = "wideo:video_usage"
    delete_url_name = "wideo:delete"
    index_url_name = "wideo:index"
    page_title = gettext_lazy("Delete video")

    def setup(self, request, *args, **kwargs):
        return super().setup(request, *args, **kwargs)

    def get_object(self, queryset=None):
        return super().get_object(queryset)

    def user_has_permission(self, permission):
        return self.permission_policy.user_has_permission_for_instance(
            self.request.user, permission, self.object
        )

    @property
    def confirmation_message(self):
        return ngettext(
            "Are you sure you want to delete this video?",
            "Are you sure you want to delete these videos?",
            1,
        )

    def get_success_message(self):
        return _("Video '%(video_title)s' deleted.") % {
            "video_title": self.object.title
        }


def add(request: HttpRequest) -> HttpResponse:
    VideoModel = get_video_model()

    if request.method == "POST":
        video = VideoModel(uploaded_by_user=request.user)
        form = BaseVideoForm(request.POST, request.FILES, instance=video)

        if form.is_valid():
            # form.save()
            obj = form.save(commit=False)
            obj.save()
            form.save_m2m()

            messages.success(
                request,
                _("Video '%(video_title)s' added.") % {"video_title": video.title},
                buttons=[
                    messages.button(reverse("wideo:edit", args=(video.id,)), _("Edit"))
                ],
            )

            return redirect("wideo:index")
        else:
            messages.error(request, _("The video could not be created due to errors."))
            form = VideoForm(request.POST, request.FILES, instance=video)
    else:
        form = VideoForm()

    return TemplateResponse(
        request,
        "wideo/videos/add.html",
        {
            "form": form,
        },
    )


@permission_checker.require("change")
def edit(request, video_id):
    Video = get_video_model()
    video = get_object_or_404(Video, id=video_id)

    if not permission_policy.user_has_permission_for_instance(
        request.user, "change", video
    ):
        raise PermissionDenied

    next_url = get_valid_next_url_from_request(request)

    if request.method == "POST":
        form_data = request.POST.copy()
        form_data["upload"] = form_data.get("upload") or video.upload_id
        form = BaseVideoForm(form_data, request.FILES, instance=video)
        if form.is_valid():
            form.save()

            edit_url = reverse("wideo:edit", args=(video.id,))
            redirect_url = "wideo:index"
            if next_url:
                edit_url = f"{edit_url}?{urlencode({'next': next_url})}"
                redirect_url = next_url

            messages.success(
                request,
                _("Video '%(video_title)s' updated.") % {"video_title": video.title},
                buttons=[messages.button(edit_url, _("Edit again"))],
            )
            return redirect(redirect_url)
        else:
            messages.error(request, _("The video could not be saved due to errors."))
            form = VideoForm(request.POST, request.FILES, instance=video)
    else:
        form = VideoForm(instance=video)

    # Check if we should enable the frontend url generator
    try:
        reverse("wideo_serve", args=("foo", "1", "bar"))
        url_generator_enabled = True
    except NoReverseMatch:
        url_generator_enabled = False

    # try:
    #     filesize = video.get_file_size()
    # except SourceVideoIOError:
    #     filesize = None

    filesize = None

    return TemplateResponse(
        request,
        "wideo/videos/edit.html",
        {
            "video": video,
            "form": form,
            # This is not supported right now
            # "url_generator_enabled": url_generator_enabled,
            "filesize": filesize,
            "user_can_delete": permission_policy.user_has_permission_for_instance(
                request.user, "delete", video
            ),
            "next": next_url,
        },
    )


@permission_checker.require("delete")
def delete(request, video_id):
    video = get_object_or_404(get_video_model(), id=video_id)

    if not permission_policy.user_has_permission_for_instance(
        request.user, "delete", video
    ):
        raise PermissionDenied

    next_url = get_valid_next_url_from_request(request)

    if request.method == "POST":
        video.delete()
        messages.success(request, _("Video '{0}' deleted.").format(video.title))
        return redirect(next_url) if next_url else redirect("wideo:index")

    return TemplateResponse(
        request,
        "wideo/videos/confirm_delete.html",
        {
            "video": video,
            "next": next_url,
        },
    )


@atomic
def upload_prepare(request: HttpRequest) -> HttpResponse:
    if request.method != "POST":
        return HttpResponse(status=status.HTTP_405_METHOD_NOT_ALLOWED)

    upload = UploadedVideo.objects.create()
    return HttpResponse(status=status.HTTP_201_CREATED, content=upload.id)


@atomic
def upload_chunk(request: HttpRequest) -> HttpResponse:
    if request.method != "PUT":
        return HttpResponse(status=status.HTTP_405_METHOD_NOT_ALLOWED)

    data, files = MultiPartParser(
        request.META, request, request.upload_handlers
    ).parse()
    upload = get_object_or_404(UploadedVideo, id=data["upload_id"])
    UploadedVideoChunk.objects.update_or_create(
        video=upload, index=data["index"], defaults={"file": files["blob"]}
    )
    return HttpResponse(status=status.HTTP_200_OK)


def get_uploaded_video_render(
    request: HttpRequest, uploaded_video_id: int
) -> HttpResponse:
    """
    Given an UploadedVideo by its ID, return its information as if it was a
    unique Render.

    This is intended to be used in the Video add/edit templates inside Wagtail
    admin, where we expect to have an UploadedVideo instance when a video file
    is uploaded but no Video instance will be created until the form is
    submitted (therefore, no Render instance will be created yet). However, we
    still want to get the information of the video to display a preview in that
    add/edit template, so we get its information as if we were getting a unique
    Render so that we can use it with a video library.
    """
    if request.method != "GET":
        return HttpResponse(status=status.HTTP_405_METHOD_NOT_ALLOWED)

    data = {
        "type": "video",
        "title": "Video preview",
        "sources": [
            {
                "src": render.file.url,
                "type": render.mime,
                "size": render.height,
            }
            for render in get_render_model().objects.filter(
                video__upload_id=uploaded_video_id
            )
        ],
        #     poster: '/path/to/poster.jpg',
        #     previewThumbnails: {
        #         src: '/path/to/thumbnails.vtt',
        #     },
        #     tracks: [
        #         {
        #             kind: 'captions',
        #             label: 'English',
        #             srclang: 'en',
        #             src: '/path/to/captions.en.vtt',
        #             default: true,
        #         },
        #         {
        #             kind: 'captions',
        #             label: 'French',
        #             srclang: 'fr',
        #             src: '/path/to/captions.fr.vtt',
        #         },
        #     ],
    }

    return HttpResponse(status=status.HTTP_200_OK, content=json.dumps(data))



# Generated by Django 4.1.9 on 2023-06-27 11:08

import django.db.models.deletion
import taggit.managers
import wagtail.models.collections
import wagtail.search.index
from django.conf import settings
from django.db import migrations, models

import wideo.storage


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("taggit", "0005_auto_20220424_2025"),
        ("wagtailcore", "0078_referenceindex"),
    ]

    operations = [
        migrations.CreateModel(
            name="UploadedVideo",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        auto_now_add=True, db_index=True, verbose_name="created at"
                    ),
                ),
                (
                    "file",
                    models.FileField(
                        help_text="The uploaded video file",
                        upload_to=wideo.storage.upload_to,
                        verbose_name="file",
                    ),
                ),
                (
                    "mime",
                    models.CharField(
                        help_text="MIME type of the video",
                        max_length=100,
                        verbose_name="mime type",
                    ),
                ),
                (
                    "duration",
                    models.FloatField(
                        help_text="Duration of the video in seconds",
                        verbose_name="duration in seconds",
                    ),
                ),
                (
                    "width",
                    models.IntegerField(
                        help_text="Width of the video in pixels",
                        verbose_name="original width",
                    ),
                ),
                (
                    "height",
                    models.IntegerField(
                        help_text="Height of the video in pixels",
                        verbose_name="original height",
                    ),
                ),
                (
                    "frames_per_second",
                    models.DecimalField(
                        decimal_places=2,
                        help_text="FPS of the video",
                        max_digits=5,
                        verbose_name="original FPS",
                    ),
                ),
                (
                    "frame_count",
                    models.IntegerField(
                        help_text="Number of frames in the video",
                        verbose_name="original frame count",
                    ),
                ),
                (
                    "uploaded_by_user",
                    models.ForeignKey(
                        blank=True,
                        editable=False,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="uploaded by user",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Video",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        auto_now_add=True, db_index=True, verbose_name="created at"
                    ),
                ),
                ("title", models.CharField(max_length=255, verbose_name="title")),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("pending", "Pending"),
                            ("processing", "Processing"),
                            ("success", "Success"),
                            ("failed", "Failed"),
                        ],
                        default="pending",
                        help_text="The status of the video processing (pending means nothing happens on this video yet, processing means the task is running, success means a render is available and failed means the task failed)",
                        max_length=10,
                        verbose_name="status",
                    ),
                ),
                (
                    "collection",
                    models.ForeignKey(
                        default=wagtail.models.collections.get_root_collection_id,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="wagtailcore.collection",
                        verbose_name="collection",
                    ),
                ),
                (
                    "tags",
                    taggit.managers.TaggableManager(
                        blank=True,
                        help_text=None,
                        through="taggit.TaggedItem",
                        to="taggit.Tag",
                        verbose_name="tags",
                    ),
                ),
                (
                    "upload",
                    models.ForeignKey(
                        help_text="The uploaded video file",
                        on_delete=django.db.models.deletion.PROTECT,
                        to="wideo.uploadedvideo",
                        verbose_name="file",
                    ),
                ),
                (
                    "uploaded_by_user",
                    models.ForeignKey(
                        blank=True,
                        editable=False,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="uploaded by user",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=(wagtail.search.index.Indexed, models.Model),
        ),
        migrations.CreateModel(
            name="Render",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        auto_now_add=True, db_index=True, verbose_name="created at"
                    ),
                ),
                (
                    "file",
                    models.FileField(
                        help_text="The uploaded video file",
                        upload_to=wideo.storage.upload_to,
                        verbose_name="file",
                    ),
                ),
                (
                    "mime",
                    models.CharField(
                        help_text="MIME type of the video",
                        max_length=100,
                        verbose_name="mime type",
                    ),
                ),
                (
                    "duration",
                    models.FloatField(
                        help_text="Duration of the video in seconds",
                        verbose_name="duration in seconds",
                    ),
                ),
                (
                    "width",
                    models.IntegerField(
                        help_text="Width of the video in pixels",
                        verbose_name="original width",
                    ),
                ),
                (
                    "height",
                    models.IntegerField(
                        help_text="Height of the video in pixels",
                        verbose_name="original height",
                    ),
                ),
                (
                    "frames_per_second",
                    models.DecimalField(
                        decimal_places=2,
                        help_text="FPS of the video",
                        max_digits=5,
                        verbose_name="original FPS",
                    ),
                ),
                (
                    "frame_count",
                    models.IntegerField(
                        help_text="Number of frames in the video",
                        verbose_name="original frame count",
                    ),
                ),
                (
                    "video",
                    models.ForeignKey(
                        help_text="The video that was rendered",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="wideo.video",
                        verbose_name="video",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
    ]



from django.core.management import BaseCommand
from django.utils.translation import gettext as _

from ... import get_video_model
from ...ffmpeg import encode_video
from ...models import AbstractVideo


class Command(BaseCommand):
    help = _("Encodes any video that has not yet been encoded")

    def handle(self, *args, **options):
        for video_id in (
            get_video_model()
            .objects.filter(status=AbstractVideo.ProcessStatus.pending)
            .values_list("id", flat=True)
        ):
            encode_video(video_id)



from django.urls import include, path, reverse
from wagtail import hooks
from wagtail.admin.menu import MenuItem

from . import admin_urls


@hooks.register("register_icons")
def register_icons(icons: list[str]) -> list[str]:
    return icons + [f"{__package__}/icons/video.svg"]


@hooks.register("register_admin_urls")
def register_admin_urls() -> list:
    return [
        path(f"{__package__}/", include(admin_urls, namespace="wideo")),
    ]


@hooks.register("register_admin_menu_item")
def register_videos_menu_item() -> MenuItem:
    return MenuItem("Videos", reverse("wideo:index"), icon_name="video")
